//
//	ADDefaults.m
//	  Written by Carl Lindberg Copyright 1999 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADDefaults.h"
#import <stdio.h>
#import <Foundation/NSUserDefaults.h>
#import <Foundation/NSProcessInfo.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSUtilities.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSException.h>
#import "ADUtilities.h"
#import "ADOutputModule.h"

NSString *ADCopyrightOwner     = @"ADCopyrightOwner";
NSString *ADOutputDirectory    = @"ADOutputDirectory";
NSString *ADProjectDirectory   = @"ADProjectDirectory";
NSString *ADParseSingleHeaders = @"ADParseSingleHeaders";
NSString *ADPreserveTree       = @"ADPreserveTree";
NSString *ADForceOverwrite     = @"ADForceOverwrite";
NSString *ADShowTimestamp      = @"ADShowTimestamp";
NSString *ADRunSilently        = @"ADRunSilently";
NSString *ADCopyDocumentation  = @"ADCopyDocumentation";
NSString *ADDebugLevel         = @"ADDebugLevel";
NSString *ADTemplateFilename   = @"ADTemplateFilename";
NSString *ADDocumentAllClasses = @"ADDocumentAllClasses";
NSString *ADCoalesceCategories = @"ADCoalesceCategories";
NSString *ADUseDocTypeDirectories = @"ADUseDocTypeDirectories";
NSString *ADAcceptStandardCComments   = @"ADAcceptStandardCComments";
NSString *ADAcceptDoubleSlashComments = @"ADAcceptDoubleSlashComments";

void usage(NSString *error)
{
	if (error) fprintf(stderr, "Error: %s\n", [error cString]);

	fprintf(stderr, "Usage: %s [options] [ sourcefiles ...]\n"
"   Options:\n"
"   -force             Force document creation for up to date files.\n"
"   -copydocs          Copy documentation files found in project tree.\n"
"   -tree              Create a tree of documentation files.\n"
"   -typedirs          Put files in Classes/Protocols/etc. subdirectories\n"
"   -nosingles         Exclude unpaired \".h\" and \".m\" files from documentation\n"
"   -timestamp         Insert a timestamp into the documentation\n"
"   -allclasses        Documents every class/category/protocol found in header\n"
"   -combine           combine category methods into main class/first category\n"
"   -help              Show this help and exit.\n"
"   -version           Show AutoDoc version/copyright and exit.\n"
"   -silent            Operate silently;  do not generate log entries.\n"
"   -Debug #           Set debugging level to #; 0 is least, 5 is most.\n"
"   -copyright \"...\"   Set copyrightowner in the copyright line to \"...\".\n"
"   -comment standardc|doubleslash   Accept other kinds of comments.\n"
"   -destination <dir> Create documentation files in directory <dir>.\n"
"   -template <file>   Use <file> as the output template\n"
"   -format rtf|html   Generate documention in either RTF or HTML format\n"
"   -project <dir>     Look for source files in the directory <dir>.\n",

	[[[NSProcessInfo processInfo] processName] cString]);

	if (error) exit(1);
	exit(0);
}

BOOL ADBoolForDefault(NSString *defaultName) {
	return [[NSUserDefaults standardUserDefaults] boolForKey:defaultName];
}
int ADIntForDefault(NSString *defaultName) {
	return [[NSUserDefaults standardUserDefaults] integerForKey:defaultName];
}
NSString *ADStringForDefault(NSString *defaultName) {
	return [[NSUserDefaults standardUserDefaults] stringForKey:defaultName];
}


NSArray *ADHandleArguments(int argc, char *argv[])
{
	NSArray *arguments = [[NSProcessInfo processInfo] arguments];
	NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
	NSMutableDictionary *argDomainDict;
	NSMutableDictionary *argDict = [NSMutableDictionary dictionary];
	NSEnumerator *argEnum = [arguments objectEnumerator];
	NSMutableArray *filesToProcess = [NSMutableArray array];
	NSString *currArg;
	NSString *copyrightOwner = nil;
	NSString *projectDirectory = nil;

	/* Set up defaults */
	[ADOutputModule setCurrentModuleFromString:@"rtf"];
	[argDict setObject:@"YES" forKey:ADParseSingleHeaders];

	[argEnum nextObject]; //skip over argv[0]
	
	while ((currArg = [argEnum nextObject]))
	{
		NSString *paramString = nil;
		NSString *fileToParse = nil;
		NSString *origArg;

/*
 * If the argument was like -flag=value, the next "argument" should be the
 * value part, not the next object from the enumerator.
 */
#define NEXT_ARG() (paramString? paramString : [argEnum nextObject])

		/* Check for -format=html type stuff */
		if ([currArg hasPrefix:@"-"])
		{
            NSRange equalsRange = [currArg rangeOfString:@"="];
			if (equalsRange.length > 0)
            {
                paramString = [currArg substringFromIndex:NSMaxRange(equalsRange)];
				currArg = [currArg substringToIndex:equalsRange.location];
				if ([paramString length] == 0) paramString = nil;
            }
        }

		origArg = currArg;

		/* Allow GNU -- style arguments as well */
		if ([currArg hasPrefix:@"--"])
			currArg = [currArg substringFromIndex:1];

		if ([@"-project" hasPrefix:currArg])
		{
			projectDirectory = ADNormalizePath(NEXT_ARG());
			[argDict setObject:projectDirectory
                                 forKey:ADProjectDirectory];
		}
		else if ([@"-destination" hasPrefix:currArg])
		{
                  NSString *path, *na;
                  na = NEXT_ARG();
                  path = ADNormalizePath(na);
                  if (path == nil) {
                    NSLog(@"couldn't normalize path '%@'", na);
                    path = na;
                  }

                  [argDict setObject:path forKey:ADOutputDirectory];
		}
		else if ([@"-copyright" hasPrefix:currArg])
		{
			copyrightOwner = NEXT_ARG();
		}
		else if ([@"-timestamp" hasPrefix:currArg])
		{
			[argDict setObject:@"YES" forKey:ADShowTimestamp];
		}
		else if ([@"-nosingles" hasPrefix:currArg])
		{
			[argDict setObject:@"NO" forKey:ADParseSingleHeaders];
		}
		else if ([@"-copydocs" hasPrefix:currArg] || [currArg isEqual:@"-rtf"])
		{
			[argDict setObject:@"YES" forKey:ADCopyDocumentation];
		}
		else if ([@"-force" hasPrefix:currArg])
		{
			[argDict setObject:@"YES" forKey:ADForceOverwrite];
		}
		else if ([@"-tree" hasPrefix:currArg])
		{
			[argDict setObject:@"YES" forKey:ADPreserveTree];
		}
		else if ([@"-format" hasPrefix:currArg])
		{
			[ADOutputModule setCurrentModuleFromString:NEXT_ARG()];
		}
		else if ([@"-silent" hasPrefix:currArg])
		{
			[argDict setObject:@"YES" forKey:ADRunSilently];
		}
		else if ([@"-allclasses" hasPrefix:currArg])
		{
			[argDict setObject:@"YES" forKey:ADDocumentAllClasses];
		}
		else if ([@"-comment" hasPrefix:currArg])
		{
			NSString *style = [NEXT_ARG() lowercaseString];

			if ([@"standardc" hasPrefix:style])
			{
				[argDict setObject:@"YES" forKey:ADAcceptStandardCComments];
			}
			else if ([@"doubleslash" hasPrefix:style])
			{
				[argDict setObject:@"YES" forKey:ADAcceptDoubleSlashComments];
			}
			//else unknown style
		}
		else if ([@"-traditional" hasPrefix:currArg])  // OBSOLETE
		{
            ADLog(@"Warning: '-traditional' is obsolete; use '-comment standardc'.");
			[argDict setObject:@"YES" forKey:ADAcceptStandardCComments];
		}
		else if ([@"-doubleslash" hasPrefix:currArg])  // OBSOLETE
		{
            ADLog(@"Warning: '-doubleslash' is obsolete; use '-comment doubleslash'.");
			[argDict setObject:@"YES" forKey:ADAcceptDoubleSlashComments];
		}
        else if ([@"-combine" hasPrefix:currArg])
        {
            [argDict setObject:@"YES" forKey:ADCoalesceCategories];
        }
        else if ([@"-typedirs" hasPrefix:currArg])
        {
            [argDict setObject:@"YES" forKey:ADUseDocTypeDirectories];
        }
		else if ([@"-template" hasPrefix:currArg])
		{
			[argDict setObject:NEXT_ARG() forKey:ADTemplateFilename];
		}
		else if ([@"-file" hasPrefix:currArg])  // to allow filenames that start with dashes
		{
			fileToParse = ADNormalizePath(NEXT_ARG());
		}
		else if ([@"-Debug" hasPrefix:currArg])
		{
			[argDict setObject:NEXT_ARG() forKey:ADDebugLevel];
		}
		else if ([@"-version" hasPrefix:currArg])
		{
			printf("AutoDoc version 2.0beta5.  Use governed by the MiscKit license.\n");
			exit(0);
		}
		else if ([@"-help" hasPrefix:currArg])
		{
			usage(nil);
		}
		else if ([currArg hasPrefix:@"-"])
		{
			usage([NSString stringWithFormat:@"unknown option: %@", origArg]);
 		}
		else
		{
			fileToParse = ADNormalizePath(currArg);
		}
	
		if (fileToParse != nil && ![filesToProcess containsObject:fileToParse])
				[filesToProcess addObject:fileToParse];

	}

	if (projectDirectory == nil && [filesToProcess count] == 0)
	{
		usage(@"No input files specified");
	}


	// If copyright wasn't on the command line, check the environment variable
	if (copyrightOwner == nil)
		copyrightOwner = [[[NSProcessInfo processInfo] environment] objectForKey:@"AD_COPYRIGHT"];
	
	if (copyrightOwner) [argDict setObject:copyrightOwner forKey:ADCopyrightOwner];

	/* Replace the NSArgumentDomain values with ours */
	argDomainDict = [[userDefaults volatileDomainForName:NSArgumentDomain] mutableCopy];
        NSCAssert(argDomainDict, @"missing argument domain dict ..");
	[argDomainDict addEntriesFromDictionary:argDict];
        NSCAssert(userDefaults, @"missing user defaults ..");
	[userDefaults removeVolatileDomainForName:NSArgumentDomain];
	[userDefaults setVolatileDomain:argDomainDict forName:NSArgumentDomain];
#if LIB_FOUNDATION_LIBRARY
        /*
          remove volatile domain currently removes the domain name
          from the search list, so add it afterwards ..
        */
        if (![[userDefaults searchList] containsObject:NSArgumentDomain]) {
          NSMutableArray *marray;
          marray = [[userDefaults searchList] mutableCopy];
          [marray insertObject:NSArgumentDomain atIndex:0];
          [userDefaults setSearchList:AUTORELEASE([marray copy])];
          RELEASE(marray);
        }
#endif
	ADDebugLog(3, @"Argument dictionary: %@", argDomainDict);
	[argDomainDict release];

	return filesToProcess;
}

