//
//	ADUtilities.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADUtilities.h"
#import "ADDefaults.h"
#import <stdio.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSCharacterSet.h>
#import <Foundation/NSUtilities.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSBundle.h>
#import <Foundation/NSArray.h>
#ifdef COCOA
#if !GNUSTEP_BASE_LIBRARY
#import <Foundation/NSCalendarDate.h>
#endif
#endif
#import "NSFileManager+Autodoc.h"
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADOutputModule.h"

/* Logs to stdout */
void ADLogv(NSString *format, va_list arguments)
{
	NSString *logString = [[NSString alloc] initWithFormat:format arguments:arguments];

	fwrite([logString cString], 1, [logString cStringLength], stdout);
	if (![logString hasSuffix:@"\n"]) fputc('\n', stdout); //hrm

	[logString release];
}

void ADLog(NSString *format, ...)
{
	va_list arguments;

	va_start(arguments, format);
	ADLogv(format, arguments);
	va_end(arguments);
}

void ADDebugLog(int debugLevel, NSString *format, ...)
{
    if (debugLevel <= ADIntForDefault(ADDebugLevel))
    {
		va_list arguments;
 
		va_start(arguments, format);
		ADLogv(format, arguments);
		va_end(arguments);
    }
}

/*
 * Our Flex/bison files "#define fprintf adfprintf", with the result being
 * that I can direct everything to stdout.
 */
void adfprintf(FILE *fp, const char *format, ...)
{
	va_list arguments;

	va_start(arguments, format);
	vfprintf(stdout, format, arguments);
	va_end(arguments);
}

#define ADD_ARRAY(array)             \
{                                    \
	NSArray *theArray = array;       \
	int i, count = [theArray count]; \
	for (i=0; i<count; i++)          \
	{                                \
		[searchPath addObject:[[theArray objectAtIndex:i] stringByAppendingPathComponent:@"AutoDoc"]]; \
	}                                \
}

NSArray *ADTemplateSearchPath()
{
	NSMutableArray *searchPath = [NSMutableArray array];
#ifdef COCOA
	int locationMask = NSUserDomainMask|NSLocalDomainMask|NSNetworkDomainMask;
	ADD_ARRAY(NSSearchPathForDirectoriesInDomains(NSAllLibrariesDirectory, locationMask, YES));
#else
	// How to add ~/Developer/AutoDoc, /LocalDeveloper/AutoDoc ?
	ADD_ARRAY(NSStandardLibraryPaths());
#endif
	[searchPath addObject:[[NSBundle mainBundle] bundlePath]];

	return searchPath;
}

NSCharacterSet *ADVariableNameSet()
{
	static NSMutableCharacterSet *variableSet = nil;
	
	if (!variableSet)
	{
		variableSet = [[NSCharacterSet alphanumericCharacterSet] mutableCopy];
		[variableSet addCharactersInString:@"_"];
	}
	
	return variableSet;
}

NSCharacterSet *ADNonVariableNameSet()
{
	static NSMutableCharacterSet *nonVariableSet = nil;
	
	if (!nonVariableSet)
	{
		nonVariableSet = [[ADVariableNameSet() invertedSet] retain];
	}
	
	return nonVariableSet;
}

BOOL ADIsCopyrightComment(NSString *commentString)
{
    return [commentString containsRegex:@"[$](Id|Header|Date|Revision):"];
}

BOOL ADIsAutodocComment(NSString *commentString)
{
	if ([commentString hasPrefix:@"//"])
    {
        return ADBoolForDefault(ADAcceptDoubleSlashComments);
    }

	/* Okay, it's a C-style comment */

	// If it's a comment like /****************/, ignore it.
	if ([commentString containsRegex:@"\\`/[*]+/\\'"])
    {
        return NO;
    }

	/* If we're accepting all C-style comments, allow anything at this point. */
	if (ADBoolForDefault(ADAcceptStandardCComments))
		return YES;

	/* Autodoc actually allows comments like /****"  */
	return ([commentString containsRegex:@"\\`/[*]+\""] &&
		    [commentString containsRegex:@"\"[*]+/\\'"]);    
}

NSRange ADRangeOfMatchedLeftString(NSString *aString, NSString *leftChar, NSString *rightChar)
{
	NSRange searchRange = {0, [aString length] };
	NSString *combinedString = [leftChar stringByAppendingString:rightChar];
	NSCharacterSet *parenSet = [NSCharacterSet characterSetWithCharactersInString:combinedString];
	int matchCount = 0;
	NSRange foundRange;
	NSString *foundString;
	
	while (1)
	{
		foundRange = [aString rangeOfCharacterFromSet:parenSet
							options:NSBackwardsSearch
							range:searchRange];

		if (foundRange.length == 0) return foundRange;

		foundString = [aString substringWithRange:foundRange];
		
		if ([foundString isEqualToString:rightChar]) matchCount++;
		else matchCount--;
		
		if (matchCount == 0) return foundRange;
		
		searchRange.length = foundRange.location;
	}
	
	return NSMakeRange(0,0);
}

NSRange ADRangeOfMatchedLeftParen(NSString *aString)
{
    return ADRangeOfMatchedLeftString(aString, @"(", @")");
}

NSRange ADRangeOfMatchedLeftSquareBracket(NSString *aString)
{
    return ADRangeOfMatchedLeftString(aString, @"[", @"]");
}

NSRange ADRangeOfVariableBeforeIndex(NSString *string, unsigned lastIndex)
{
	NSRange searchRange = {0, lastIndex};
	NSRange nameStartRange, nameEndRange;
	NSRange variableRange;

	nameEndRange = [string rangeOfCharacterFromSet:ADVariableNameSet()
							options:NSBackwardsSearch
							range:searchRange];

	if (nameEndRange.length == 0) return NSMakeRange(NSNotFound, 0);

	searchRange.length = nameEndRange.location;
	nameStartRange = [string rangeOfCharacterFromSet:ADNonVariableNameSet()
							options:NSBackwardsSearch
							range:searchRange];
	
	if (nameStartRange.length == 0)
		variableRange.location = 0;
	else
		variableRange.location = NSMaxRange(nameStartRange);
	variableRange.length = NSMaxRange(nameEndRange) - variableRange.location;

	return variableRange;
}

/*
 * Used by typedefs, ivars, and global vars; this attempts to pluck the
 * variable name out of the fully-typed variable string.  This is generally
 * the last word in the string, but this skips over type information that
 * can follow the variable name (currently bitfields, arrays, and function
 * pointer definition stuff).
 */
NSRange ADRangeOfVariableInTypeString(NSString *string)
{
	NSRange bitfieldRange = [string rangeOfRegex:@":[0-9]+[ \t\n\r]*;"];
	NSRange arrayRange    = [string rangeOfRegex:@"([ \t\n\r]*[[][^]]*[]])+[ \t\n\r]*;"];
	NSRange funcRange     = [string rangeOfRegex:@"[)][ \t\n\r]*;"];

	if (bitfieldRange.location != NSNotFound)
	{		
		return ADRangeOfVariableBeforeIndex(string, bitfieldRange.location);
	}
	else if (arrayRange.location != NSNotFound)
	{
		return ADRangeOfVariableBeforeIndex(string, arrayRange.location);
	}
	else if (funcRange.location != NSNotFound) // typed function ivar
	{
		NSRange leftParenRange  = ADRangeOfMatchedLeftParen(string);
		return ADRangeOfVariableBeforeIndex(string, leftParenRange.location);
	}
	else
	{
		return ADRangeOfVariableBeforeIndex(string, [string length]);
	}
}

NSString *ADSquashTypeString(NSString *aString)
{
	aString = [aString stringByTrimmingLeadWhitespace];
	aString = [aString stringByReplacingEverySeriesOfCharactersFromSet:
						[NSCharacterSet characterSetWithCharactersInString:@" \t\n\r\v\f"]
						withString:@" "];
	return aString;
}

/* Could be an NSString method */
NSString *ADPathWithExtension(NSString *path, NSString *extension)
{
	return [[path stringByDeletingPathExtension] stringByAppendingPathExtension:extension];
}

NSString *ADNormalizePath(NSString *path)
{
	NSFileManager *manager = [NSFileManager defaultManager];
	if (![path isAbsolutePath])
		path = [[manager currentDirectoryPath] stringByAppendingPathComponent:path];

	return [path stringByStandardizingPath];  
}

BOOL ADFileExists(NSString *filename)
{
	return [[NSFileManager defaultManager] regularFileExistsAtPath:filename];
}

BOOL ADFileIsNewer(NSString *file1, NSString *file2)
{
	NSFileManager *manager = [NSFileManager defaultManager];
	NSDate *date1 = [[manager fileAttributesAtPath:file1 traverseLink:YES] fileModificationDate];
	NSDate *date2 = [[manager fileAttributesAtPath:file2 traverseLink:YES] fileModificationDate];

	if (date1 == nil || date2 == nil) return YES;
	return ([date1 compare:date2] == NSOrderedDescending);
}

/* May want a way to turn auto-bolding off... */
NSArray *ADBoldWordsArray()
{
	static NSArray *boldArray = nil;
	
	if (boldArray == nil)
	{
		boldArray = [[NSArray alloc] initWithObjects:@"id", @"self", @"super", @"nil", nil];
	}
	
	return boldArray;
}

/* Does the formatting for _{...} type of markup. */
static NSString *ADFormatListMarkup(NSString *string, ADOutputModule *outputModule)
{
	NSMutableString *newString = [NSMutableString stringWithCapacity:[string length]];
	unsigned stringLength = [string length];
	NSRange searchRange = NSMakeRange(0, stringLength);
	NSRange foundRange;
	NSRange betweenRange = NSMakeRange(0, 0);
	NSString *listRegex = @"_{ *([^} ]+) *([^}]*)} *";
	NSString *listGroupRegex = [NSString stringWithFormat:@"(%@[ \n\r\t]*)+", listRegex];
	
	foundRange = [string rangeOfRegex:listGroupRegex range:searchRange];
	while (foundRange.location != NSNotFound)
	{
		NSString *listsString;
		betweenRange.length = foundRange.location - betweenRange.location;
		[newString appendString:[string substringWithRange:betweenRange]];
		
		[newString appendString:[outputModule listPrefixString]];
		listsString = [string substringWithRange:foundRange];
		listsString = [listsString stringByReplacingEveryOccurrenceOfRegex:listRegex
							withString:[outputModule listItemReplaceString]
							options:MiscUseMatchSubstitutions];

		/* Reformat paragraph separators with special list versions */
		listsString = [listsString stringByReplacingEveryOccurrenceOfString:
							[outputModule paragraphSeparator]
							withString:[outputModule listParagraphSeparator]];

		[newString appendString:listsString];
		[newString appendString:[outputModule listPostfixString]];
		
		betweenRange.location = NSMaxRange(foundRange);
		searchRange = NSMakeRange(betweenRange.location, stringLength - betweenRange.location);
		foundRange = [string rangeOfRegex:listGroupRegex range:searchRange];
	}
	
	[newString appendString:[string substringFromIndex:betweenRange.location]];

	return newString;
}

/*"
 * Formats the string, converting autodoc-style markup into the appropriate
 * output format.
"*/
/*
 * Maybe it would be a good idea to have another lexer program to parse the
 * comment string... it would probably be a bit more flexible, and we could
 * maybe have different lexers to handle different kinds of markup in
 * comments in case we want to allow something other than the #{}, %{}, _{}
 * stuff we have now.
 */
NSString *ADFormatString(NSString *string, NSArray *italicWordsArray)
{
	ADOutputModule *outputModule = [ADOutputModule currentModule];
	NSArray *boldWordsArray = ADBoldWordsArray();
	NSString *startBold = [outputModule startBold];
	NSString *stopBold  = [outputModule stopBold];
	NSString *startItal = [outputModule startItalics];
	NSString *stopItal  = [outputModule stopItalics];
	NSString *boldReplace1   = [NSString stringWithFormat:@"%@$1%@",   startBold, stopBold];
	NSString *boldReplace2   = [NSString stringWithFormat:@"$1%@$2%@", startBold, stopBold];
	NSString *italicReplace1 = [NSString stringWithFormat:@"%@$1%@",   startItal, stopItal];
	NSString *italicReplace2 = [NSString stringWithFormat:@"$1%@$2%@", startItal, stopItal];
	NSEnumerator *wordEnum;
	NSString     *currWord;

	string = [outputModule preProcessString:string];

    /* boldify text of the form #{the text} */
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"#{([^}]+)}"
						withString:boldReplace1
						options:MiscUseMatchSubstitutions];
	
    /* italicize text of the form %{the text} */
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"%{([^}]+)}"
						withString:italicReplace1
						options:MiscUseMatchSubstitutions];

    /* boldify words which start with a "#" in front of them */
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"(\\`|[^\\\\])#((\\w|:|_)+)"
						withString:boldReplace2
						options:MiscUseMatchSubstitutions];

    /* italicize words which start with a "%" in front of them */
    // orig autodoc didn't have ':' as part of pattern
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"(\\`|[^\\\\])%((\\w|:|_)+)"
						withString:italicReplace2
						options:MiscUseMatchSubstitutions];


    /* boldify words in boldWordsArray */
	wordEnum = [boldWordsArray objectEnumerator];
	while ((currWord = [wordEnum nextObject]))
	{
		NSString *repl  = [NSString stringWithFormat:@"$1%@$2%@$3", startBold, stopBold];
		NSString *regex = [NSString stringWithFormat:
							@"(\\`|[^_a-zA-Z0-9])(%@)(\\'|[^_a-zA-Z0-9])", currWord];
		string = [string stringByReplacingEveryOccurrenceOfRegex:regex
							withString:repl
							options:MiscUseMatchSubstitutions];
	}
	
    /* italicize words in italicWordsArray */
	wordEnum = [italicWordsArray objectEnumerator];
	while ((currWord = [wordEnum nextObject]))
	{
		NSString *repl  = [NSString stringWithFormat:@"$1%@$2%@$3", startItal, stopItal];
		NSString *regex = [NSString stringWithFormat:
							@"(\\`|[^_a-zA-Z0-9])(%@)(\\'|[^_a-zA-Z0-9])", currWord];
		if ([currWord length] == 0) continue;   //Make sure
		string = [string stringByReplacingEveryOccurrenceOfRegex:regex
							withString:repl
							options:MiscUseMatchSubstitutions];
	}
	
    /* Reformat '\%' and '\#' and '\/' to just '%' and '#' and '/' */
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"\\\\([/#%])"
							withString:@"$1" options:MiscUseMatchSubstitutions];

	/* Make double newline markers into paragraph separators */
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"\r?\n\r?\n[\n\r\t ]*"
							withString:[outputModule paragraphSeparator]];

	/* Format _{...} type of markup */
	string = ADFormatListMarkup(string, outputModule);
	
	string = [outputModule postProcessString:string];
	
	return string;
}

/* This is just used to indent typedefs. */
NSString *ADFormatValueString(NSString *string)
{
	ADOutputModule *outputModule = [ADOutputModule currentModule];
    NSString       *indentString = @"__INDENT__";

	string = [string stringByReplacingEveryOccurrenceOfRegex:@"\r?\n[\t ]*" withString:@"\n"];
	string = [string stringByReplacingEveryOccurrenceOfRegex:@"[\t ]+" withString:@" "];
	
	if ([string containsString:@"{"])
	{
		NSString *indent = indentString;
		NSArray *lineArray = [string componentsSeparatedByString:@"\n"];
		NSEnumerator *lineEnum = [lineArray objectEnumerator];
		NSString *line;
		NSMutableArray *newArray = [NSMutableArray arrayWithCapacity:[lineArray count]];

		while ((line = [lineEnum nextObject]))
		{
			if ([line containsString:@"{"])
			{
				NSString *replace;
				indent = [indent stringByAppendingString:indentString];
				replace = [NSString stringWithFormat:@"{%@$1", indent];
				line = [line stringByReplacingRegex:@"{ *([^ ]*)"
							withString:replace
							options:MiscUseMatchSubstitutions];
			}
			else if ([line containsString:@"}"])
			{
                if ([indent length] > [indentString length])
                    indent = [indent substringFromIndex:[indentString length]];
				line = [NSString stringWithFormat:@"%@%@", indent, line];
			}
			else
			{
				line = [NSString stringWithFormat:@"%@%@", indent, line];
			}
			
			[newArray addObject:line];
		}
		
		string = [newArray componentsJoinedByString:@"\n"];
	}
	
	string = [outputModule preProcessString:string];
	string = [string stringByReplacingEveryOccurrenceOfString:@"\n"
					withString:[outputModule lineSeparator]];
    string = [string stringByReplacingEveryOccurrenceOfString:indentString
                    withString:[outputModule valueIndentString]];
	string = [outputModule postProcessString:string];
	
	return string;
}

