//
//	ADFile.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADFile.h"
#import "ADDefaults.h"
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSUtilities.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSAutoreleasePool.h>
#ifdef COCOA
#import <Foundation/NSCalendarDate.h>
#endif
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADDefaults.h"
#import "ADInlineFunction.h"
#import "ADGlobalVariable.h"
#import "ADConstant.h"
#import "ADMacro.h"
#import "ADTypedef.h"
#import "ADInterface.h"
#import "ADClass.h"
#import "ADProtocol.h"
#import "ADCategory.h"
#import "ADComment.h"
#import "ADUtilities.h"

@implementation ADFile
/*"
 * An ADFile object is a container all the information parsed from the
 * source code. It has arrays for all the possible elements that we can
 * find, plus some additional items such as the version and copyright year
 * that could be found in the code.
 * 
 * The ADFile object is the top-level object passed to the merge engine, so
 * the field names in the templates start with the methods of this class.
 * The set methods here are mainly for the parser's use, while the query
 * methods are for the templates.
"*/

/*" Initializes a new ADFile instance. "*/
- init
{
	[super init];
	
    inlineFunctionArray = [[NSMutableArray alloc] init];
    functionArray = [[NSMutableArray alloc] init];
	variableArray = [[NSMutableArray alloc] init];
	constantArray = [[NSMutableArray alloc] init];
	macroArray    = [[NSMutableArray alloc] init];
	typedefArray  = [[NSMutableArray alloc] init];

	classArray    = [[NSMutableArray alloc] init];
	categoryArray = [[NSMutableArray alloc] init];
	protocolArray = [[NSMutableArray alloc] init];

	version = @"1.1";
	copyrightYear = [[(NSCalendarDate *)[NSCalendarDate date]
									 descriptionWithCalendarFormat:@"%Y"] retain];
	[self setCopyrightOwner:ADStringForDefault(ADCopyrightOwner)];

	return self;
}

- (void)dealloc
{
    [inlineFunctionArray release];
	[functionArray release];
	[variableArray release];
	[constantArray release];
	[macroArray release];
	[typedefArray release];
	[classArray release];
	[categoryArray release];
	[protocolArray release];
	[mainInterface release];
	[version release];
	[copyrightYear release];
	[copyrightOwner release];
	[headerFilename release];
	[sourceFilename release];
	[importString release];
	[super dealloc];
}

- (NSString *)version			{	return version;			}
- (NSString *)copyrightYear		{	return copyrightYear;	}
- (NSString *)copyrightOwner	{	return copyrightOwner;	}
- (NSString *)headerFilename	{	return headerFilename;	}
- (NSString *)sourceFilename	{	return sourceFilename;	}
- (NSString *)importString		{	return importString;	}
- (id)mainInterface				{	return mainInterface;	}
- (BOOL)hasMainInterface		{	return (mainInterface != nil);	}
- (BOOL)hasParsed               {   return hasParsed;       }

/*"
 * Returns YES if there's at least one class, category, function, macro,
 * etc., or NO otherwise.  If there is nothing to document in a file, then
 * no documentation file should be generated.
"*/
- (BOOL)hasDocumentableElements
{
    if ([self hasMainInterface]) return YES;

    if ([functionArray count] > 0 || [macroArray count] > 0 ||
        [constantArray count] > 0 || [inlineFunctionArray count] > 0 ||
        [variableArray count] > 0 || [typedefArray count] > 0)
    {
        return YES;
    }

    return NO;
}

- (NSString *)mainTitle
{
	if ([self hasMainInterface])
		return [[self mainInterface] title];
	else
		return [[[self headerFilename] lastPathComponent] stringByDeletingPathExtension];
}

/*"
 * When -typedirs is given, this should return the name of the subdirectory
 * under the destination directory this file should be placed in
 * (i.e."Classes","Protocols", "TypesAndConstants", etc.).
"*/
- (NSString *)typeDirectoryName
{
    if ([self hasMainInterface])
        return [[self mainInterface] typeDirectoryName];
    else if ([functionArray count] > 0 || [inlineFunctionArray count] > 0 || [macroArray count] > 0)
        return @"Functions";
    else
        return @"TypesAndConstants";
}

- (void)setHasParsed:(BOOL)parsed               {	hasParsed = parsed;     			}
- (void)setVersion:(NSString *)aString			{	SETIVAR(version, aString);			}
- (void)setCopyrightYear:(NSString *)aString	{	SETIVAR(copyrightYear, aString);	}
- (void)setCopyrightOwner:(NSString *)aString	{	SETIVAR(copyrightOwner, aString);	}
- (void)setHeaderFilename:(NSString *)aString	{	SETIVAR(headerFilename, aString);	}
- (void)setSourceFilename:(NSString *)aString	{	SETIVAR(sourceFilename, aString);	}
- (void)setImportString:(NSString *)aString		{	SETIVAR(importString, aString);		}
- (void)setMainInterface:(ADInterface*)anInterface {SETIVAR(mainInterface, anInterface);}

- (NSArray *)inlineFunctionArray { return inlineFunctionArray; }
- (NSArray *)functionArray	{	return functionArray;	}
- (NSArray *)variableArray	{	return variableArray;	}
- (NSArray *)constantArray	{	return constantArray;	}
- (NSArray *)macroArray		{	return macroArray;		}
- (NSArray *)typedefArray	{	return typedefArray;	}

- (void)addInlineFunction:(ADInlineFunction *)aFunction {
    [inlineFunctionArray addObject:aFunction];
}
- (void)addFunction:(ADFunction *)aFunction		{	[functionArray addObject:aFunction];	}
- (void)addVariable:(ADGlobalVariable *)aVariable{	[variableArray addObject:aVariable];	}
- (void)addConstant:(ADConstant *)aConstant		{	[constantArray addObject:aConstant];	}
- (void)addMacro:(ADMacro *)aMacro				{	[macroArray addObject:aMacro];			}
- (void)addTypedef:(ADTypedef *)aTypedef		{	[typedefArray addObject:aTypedef];		}

- (void)addClass:(ADClass *)aClass;
{
	[classArray addObject:aClass];
	if (!mainInterface) mainInterface = [aClass retain];
}

- (void)addProtocol:(ADProtocol *)aProtocol;
{
	[protocolArray addObject:aProtocol];
	if (!mainInterface) mainInterface = [aProtocol retain];
}

- (void)addCategory:(ADCategory *)aCategory;
{
	[categoryArray addObject:aCategory];
	if (!mainInterface) mainInterface = [aCategory retain];
}
- (void)addInterface:(ADInterface *)interface
{
	if ([interface isKindOfClass:[ADClass class]]) [self addClass:(id)interface];
	if ([interface isKindOfClass:[ADCategory class]]) [self addCategory:(id)interface];
	if ([interface isKindOfClass:[ADProtocol class]]) [self addProtocol:(id)interface];
}

- (ADClass *)classNamed:(NSString *)className
{
	NSEnumerator *classEnum = [classArray objectEnumerator];
	ADClass		*currClass;
	
	while ((currClass = [classEnum nextObject]))
	{
		if ([[currClass className] isEqualToString:className])
			return currClass;
	}
	
	return nil;
}


- (ADProtocol *)protocolNamed:(NSString *)protocolName
{
	NSEnumerator *protocolEnum = [protocolArray objectEnumerator];
	ADProtocol	*currProtocol;
	
	while ((currProtocol = [protocolEnum nextObject]))
	{
		if ([[currProtocol protocolName] isEqualToString:protocolName])
			return currProtocol;
	}
	
	return nil;
}

- (ADCategory *)categoryNamed:(NSString *)categoryName forClass:(NSString *)className
{
	NSEnumerator *categoryEnum = [categoryArray objectEnumerator];
	ADCategory	*currCategory;
	
	while ((currCategory = [categoryEnum nextObject]))
	{
		if ([[currCategory categoryName] isEqualToString:categoryName] &&
			 [[currCategory className] isEqualToString:className])
		{
			return currCategory;
		}
	}
	
	return nil;
}

/*"
 * Searches copyrightString for version and copyright information inside of
 * RCS tags, and sets them if found.  Mainly for use by the bison parser.
"*/
- (void)setCopyrightFromString:(NSString *)copyrightString
{
	NSArray *strings;
	
	strings = [copyrightString
					grep:@"\\$(Header|Id):[ \t]+[^ \t]+[ \t]+([^ \t]+)[ \t]+([0-9]+)/"
					forPiece:MiscSubstringsPiece];

	if ([strings count] >= 4)
	{
		[self setVersion:[strings objectAtIndex:2]];
		[self setCopyrightYear:[strings objectAtIndex:3]];
	}

	strings = [copyrightString grep:@"\\$Date:[ \t]+([0-9]+)/" forPiece:MiscSubstringsPiece];
	if ([strings count] >= 2)
    {
        [self setCopyrightYear:[strings objectAtIndex:1]];
    }

	strings = [copyrightString grep:@"\\$Revision:[ \t]+([^ \t]+)"
					forPiece:MiscSubstringsPiece];
	if ([strings count] >= 2)
    {
        [self setVersion:[strings objectAtIndex:1]];
    }
}

/*"
 * Sees if aString appears to be the main #import statement for this file
 * (i.e. contains the header filename), and if so, sets the importString
 * ivar accordingly.  For use by the bison parser.
"*/
- (void)setImportFromString:(NSString *)aString
{
	NSString *regex = [NSString stringWithFormat:@"[\"<](.*%@)[\">]",
										[headerFilename lastPathComponent]];
	NSArray *strings = [aString grep:regex forPiece:MiscSubstringsPiece];
	
	if ([strings count] >= 2)
	{
		[self setImportString:[[strings objectAtIndex:1] stringByTrimmingWhitespace]];
	}
}

/*"
 * Support method for creating the array of subarrays of header entitys
 * (i.e. typedefs, macros, etc.).  ADFile stores these entities in a single
 * array, but often only subsets of those should be grouped together with
 * the same description in the generated documentation.  This method takes
 * the base array for any of the header entities, and breaks them up into
 * an array of NSDictionaries, where each dictionary has two keys:
 * "comment", which holds the ADComment for the group, and "itemArray",
 * which is the subarray of entities to be grouped together in the
 * documentation.
 * 
 * This method is mainly a utility method for use by -#macroGroupArray, etc.
 * and should not be called directly.  However, the key names in the
 * dictionaries are used in the template files.
"*/
- (NSArray *)groupArrayFromEntityArray:(NSArray *)entityArray
{
	NSMutableArray	*returnArray = [NSMutableArray array];
	NSEnumerator	*entityEnum = [entityArray objectEnumerator];
	ADHeaderEntity	*currEntity;
	NSMutableArray  *currItemArray = nil;
	NSMutableDictionary *currGroup;

	while ((currEntity = [entityEnum nextObject]))
	{
		if (![currEntity relatedToPreviousEntity] || currItemArray == nil)
		{
			currGroup = [NSMutableDictionary dictionary];
			currItemArray = [NSMutableArray array];
			[currGroup setObject:currItemArray forKey:@"itemArray"];
			if ([currEntity entityDescription])
				[currGroup setObject:[currEntity entityDescription] forKey:@"comment"];
			[returnArray addObject:currGroup];
		}

		[currItemArray addObject:currEntity];
	}
	
	return returnArray;
}

- (NSArray *)inlineFunctionGroupArray
{	return [self groupArrayFromEntityArray:inlineFunctionArray];
}
- (NSArray *)functionGroupArray
{	return [self groupArrayFromEntityArray:functionArray];
}
- (NSArray *)variableGroupArray
{	return [self groupArrayFromEntityArray:variableArray];
}
- (NSArray *)constantGroupArray
{	return [self groupArrayFromEntityArray:constantArray];
}
- (NSArray *)macroGroupArray
{	return [self groupArrayFromEntityArray:macroArray];
}
- (NSArray *)typedefGroupArray
{	return [self groupArrayFromEntityArray:typedefArray];
}

/*"
 * Returns the a string representing the current time and date, suitable
 * for use as the timestamp line in the generated documentation.  The
 * calendar format is currently " %a %b %d %H:%M:%S %Y".
"*/
- (NSString *)timestampString
{
	if (!ADBoolForDefault(ADShowTimestamp)) return @"";
	return [(id)[NSCalendarDate date] descriptionWithCalendarFormat:@" %a %b %d %H:%M:%S %Y"];
}

/*"
 * Returns an array of all the classes, categories, and protocols contained
 * in the receiver.
"*/
- (NSArray *)interfaceArray;
{
	return [[classArray arrayByAddingObjectsFromArray:categoryArray]
			arrayByAddingObjectsFromArray:protocolArray];
}

/*"
 * Returns an array of ADFile objects; one for each class, category, and
 * protocol found in the receiver.  The first object in the returned array
 * will be the receiver itself.  This method is used to help implement the
 * -allclasses flag; when that is used, the merge engine is called once for
 * each ADFile in the array returned from this method.
"*/
- (NSArray *)subfileArray
{
	NSMutableArray *fileArray = [NSMutableArray array];
	NSEnumerator   *interfaceEnum = [[self interfaceArray] objectEnumerator];
	ADInterface    *currInterface;

	[fileArray addObject:self];

	while ((currInterface = [interfaceEnum nextObject]))
	{
		if (currInterface != mainInterface)
		{
			ADFile *newFile = [[[ADFile alloc] init] autorelease];

			[newFile setImportString:[self importString]];
			[newFile setSourceFilename:[self sourceFilename]];
			[newFile setHeaderFilename:[self headerFilename]];
			[newFile setCopyrightOwner:[self copyrightOwner]];
			[newFile setCopyrightYear:[self copyrightYear]];
			[newFile setVersion:[self version]];

			[newFile addInterface:currInterface];
			[fileArray addObject:newFile];
		}
	}

	return fileArray;
}

- (ADInterface *)mainInterfaceForCategory:(ADCategory *)category
{
    NSString    *className = [category className];
    ADInterface *interface = [self classNamed:className];

    if (interface == nil)
    {
        int i, count = [categoryArray count];

        for (i=0; i<count; i++)
        {
            ADCategory *currCategory = [categoryArray objectAtIndex:i];

            if (category == currCategory)
                break;

            if ([[currCategory className] isEqualToString:className])
            {
                interface = currCategory;
                break;
            }
        }
    }

    return interface;
}

/*"
 * Combines all the methods in categories into their ADClass if it's
 * defined in the file, or otherwise into the first ADCategory defined for
 * that class.  Used for the -combine command line option.
"*/
- (void)coalesceCategories
{
    NSMutableArray *removedCategories = [NSMutableArray array];
    NSEnumerator   *categoryEnum = [categoryArray objectEnumerator];
    ADCategory     *currCategory;
    ADInterface    *topInterface;

    while ((currCategory = [categoryEnum nextObject]))
    {
      if ((topInterface = [self mainInterfaceForCategory:currCategory]))
        {
            [topInterface addDeclarationsFromArray:[currCategory declarationArray]];
            [removedCategories addObject:currCategory];
        }
    }

    [categoryArray removeObjectsInArray:removedCategories];
}

static void _addEntities(NSMutableArray *descArray, NSArray *entityArray, NSString *entityName)
{
	if ([entityArray count] > 0)
	{
		NSEnumerator *entityEnum = [entityArray objectEnumerator];
		ADEntity	 *currEntity;
		
		[descArray addObject:entityName];
		
		while ((currEntity = [entityEnum nextObject]))
		{
			if ([currEntity entityDescription] != nil)
				[descArray addObject:[[currEntity entityDescription] description]];
			[descArray addObject:[currEntity description]];
		}
	}
}

- description
{
	NSMutableArray *descArray = [NSMutableArray array];
	
	[descArray addObject:
		[NSString stringWithFormat:@"Version %@ Copyright %@ by %@. All Rights Reserved. %@",
			version,
			copyrightYear,
			copyrightOwner,
			[self timestampString]]];
	
	[descArray addObject:[NSString stringWithFormat:@"Declared In: %@", importString]];

	if (mainInterface)
		[descArray addObject:[[mainInterface title] uppercaseString]];
	else
		[descArray addObject:@"UTILITIES"];
	
	_addEntities(descArray, typedefArray,  @"Defined Types");
	_addEntities(descArray, constantArray, @"Symbolic Constants");
	_addEntities(descArray, macroArray,    @"Macro Definitions");
    _addEntities(descArray, functionArray, @"Functions");
    _addEntities(descArray, inlineFunctionArray, @"Inline Functions");
	_addEntities(descArray, variableArray, @"Global Variables");

//	[descArray addObject:[classArray description]];
//	[descArray addObject:[protocolArray description]];
//	[descArray addObject:[categoryArray description]];

	if (mainInterface) [descArray addObject:[mainInterface description]];
	
	return [descArray componentsJoinedByString:@"\n"];
}

@end

