//
//	ADIvar.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADIvar.h"
#import <Foundation/NSArray.h>
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADCompoundIvar.h"
#import "ADComment.h"
#import "ADUtilities.h"

@implementation ADIvar

/*" This class represents a single instance variable. "*/

/*" Returns a new, autoreleased ADIvar with only the name portion set. "*/
+ ivarWithName:(NSString *)aName
{
	return [[[self alloc] initWithName:aName] autorelease];
}

/*" Returns a new, autoreleased ADIvar. "*/
+ ivarWithString:(NSString *)string
{
	return [[[self alloc] initWithString:string] autorelease];
}

/*" Initializes the ADIvar with the name portion only. "*/
- initWithName:(NSString *)aName
{
	[super init];
	name = [aName retain];
	type = @"";
	posttype = @"";
	[self setDescriptionString:@"No description."];
	return self;
}

/*"
 * This is the designated initializer for ADIvar.  Parses string (which
 * should contain the complete declaration from the header), and pulls out
 * the type, name, and (if present) posttype portions.  If string contains
 * "INTERNAL", returns nil, otherwise, returns self.
 * 
 * This "INTERNAL" feature, which caused ivars so marked to not be listed
 * in the generated documentation, was undocumented in the original
 * autodoc.  Perhaps the check should be done in the grammar file instead,
 * but it's here for now.
"*/
- initWithString:(NSString *)string
{
	NSRange nameRange;
	NSRange semicolonRange;
	NSRange suffixRange;
	
	[super init];

    /*
     * Undocumented AutoDoc feature -- you can mark ivars as INTERNAL and
     * it won't document them.
     */
	if ([string containsString:@"INTERNAL"])
	{
		[self release];
		return nil;
	}

	nameRange      = ADRangeOfVariableInTypeString(string);
	semicolonRange = [string rangeOfString:@";" options:NSBackwardsSearch];
	suffixRange.location = NSMaxRange(nameRange);
	suffixRange.length   = semicolonRange.location - suffixRange.location;
	
	type     = ADSquashTypeString([string substringToIndex:nameRange.location]);
	name     = [string substringWithRange:nameRange];
	posttype = [[string substringWithRange:suffixRange] stringBySquashingWhitespace];

	[name retain];
	[type retain];
	[posttype retain];

	[self setDescriptionString:@"No description."];

	return self;
}

- (void)dealloc
{
	[name release];
	[type release];
	[super dealloc];
}
	
- (NSString *)name
{
	return name;
}

/*"
 * Returns any type information that comes after the variable name, such as
 * bitfield or array sizes.
"*/
- (NSString *)posttype
{
	return posttype;
}

- (NSString *)type
/*" Returns the ivar's type. "*/
{
	return type;
}

/*" Returns the ADCompound the receiver is part of, if any. "*/
- (ADCompoundIvar *)parentIvar
{
	return parentIvar;
}

/*" Sets the receiver to be a subIvar of anIvar.  anIvar is not retained. "*/
- (void)setParentIvar:(ADCompoundIvar *)anIvar
{
	parentIvar = anIvar;
}


/*"
 * Sets the receiver's type description.
"*/
- (void)setType:(NSString *)aType
{
	SETIVAR(type, aType);
}

/*"
 * Returns a string representing the ivar's full name, suitable for use in
 * the documentation when documenting the individual ivars.  Normally this
 * is just the ivar's name, but for ivars which are part of a union or
 * struct, this will return a "dot path" with the parentIvar's name coming
 * first.
"*/
- (NSString *)fullName
{
	if (parentIvar)
		return [NSString stringWithFormat:@"%@.%@", [parentIvar fullName], name];
	else
		return name;
}

/*" Returns a description of the ivar. "*/
- description
{
	return [NSString stringWithFormat:@"%@|%@|%@; %@", type, name, posttype, entityDescription];
}


@end
