//
//	ADMethod.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADMethod.h"
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADComment.h"
#import "ADUtilities.h"

@implementation ADMethod
/*"
 * ADMethod represents a single Objective-C method.  It breaks each method
 * up into the argumentNames, argumentTypes, and variableNames arrays,
 * where each is an NSArray of NSStrings.  argumentNames holds the actual
 * method name components (minus the semicolons), argumentTypes holds the
 * type of each argument (a blank string if there was no declared type),
 * and variableNames holds the variable names of the arguments.  If there
 * are no parameters for the method, the argumentNames array will have a
 * single NSString which is the method name, and the other two arrays will
 * be empty.  The isClassMethod BOOL ivar tells whether the receiver is a
 * class method or instance method.
 * 
 * For the declaration "- (void)replaceObjectAtIndex:(unsigned)anIndex
 * withObject:anObject", the argument names will have
 * ("replaceObjectAtIndex", "withObject"), the argumentTypes will have
 * ("(unsigned)", ""), variableNames will have ("anIndex", "anObject"),
 * isClassMethod will be NO, and returnType will be "(void)".
 * 
 * For the declaration "- description", the argument names will have
 * ("description"), the argumentTypes and variableNames will be empty
 * arrays, isClassMethod will be NO, and returnType will be "".
"*/

+ method
{
	return [[[self alloc] init] autorelease];
}

- init
{
	[super init];
	argumentNames = [[NSMutableArray alloc] init];
	argumentTypes = [[NSMutableArray alloc] init];
	variableNames = [[NSMutableArray alloc] init];
	[self setDescriptionString:@"No method description."];
	return self;
}

- (void)dealloc
{
	[returnType release];
	[argumentNames release];
	[argumentTypes release];
	[variableNames release];
	[super dealloc];
}

- (BOOL)isClassMethod
{
	return isClassMethod;
}

- (void)setClassMethod:(BOOL)yorn
{
	isClassMethod = yorn;
}
- (BOOL)isVarargMethod
{
	return isVarargMethod;
}
- (void)setIsVarargMethod:(BOOL)yorn
{
	if ([self argumentCount] > 0)
	{
		isVarargMethod = yorn;

		if (isVarargMethod)
		{
			NSString *lastName = [variableNames lastObject];
			NSString *newName = [NSString stringWithFormat:@"%@, ...", lastName];
			unsigned lastIndex = [variableNames count] - 1;
			[variableNames replaceObjectAtIndex:lastIndex withObject:newName];
		}
	}
}

- (NSString *)returnType
{
	return returnType;
}

- (void)setReturnType:(NSString *)type
{
	SETIVAR(returnType, type);
}

/*"
 * A method used by the parser.  It adds another argument name, type, and
 * variable name as they are parsed.
"*/
- (void)addArgument:(NSString *)argName type:(NSString *)argType varName:(NSString *)varName
{
	[argumentNames addObject:argName];
	[argumentTypes addObject:argType];
	[variableNames addObject:varName];
}

/*"
 * A method used by the parser.  It handles the situation when there are no
 * parameters for a method.
"*/
- (void)setFirstArgumentName:(NSString *)aName
{
	if ([self argumentCount] > 0)
		[argumentNames replaceObjectAtIndex:0 withObject:aName];
	else
		[argumentNames addObject:aName];
}

- (NSArray *)argumentNames
{
	return argumentNames;
}

- (NSArray *)argumentTypes
{
	return argumentTypes;
}

- (NSArray *)variableNames
{
	return variableNames;
}

- (NSArray *)arguments
{
	int i, count = [self argumentCount];
	NSMutableArray *argumentArray = [NSMutableArray arrayWithCapacity:count];
		
	for (i=0;i<count;i++)
	{
		NSMutableDictionary *argDict = [NSMutableDictionary dictionaryWithCapacity:3];
		
		[argDict setObject:[argumentNames objectAtIndex:i] forKey:@"argName"];
		[argDict setObject:[argumentTypes objectAtIndex:i] forKey:@"argType"];
		[argDict setObject:[variableNames objectAtIndex:i] forKey:@"varName"];
		
		[argumentArray addObject:argDict];
	}

	return argumentArray;
}

/*"
 * Returns a string describing the method which only contains the argument
 * names.  Basically the same thing as a selector name.
"*/
- (NSString *)selectorName
{
	NSMutableString *returnString = [NSMutableString stringWithCString:""];
		
	if ([self argumentCount] == 0)
	{
		[returnString appendString:[argumentNames objectAtIndex:0]];
	}
	else
	{
		int i, count = [self argumentCount];
		
		for (i=0;i<count;i++)
		{
			[returnString appendFormat:@"%@:",[argumentNames objectAtIndex:i]];
		}
	}
	
	return returnString;
}

/*"
 * Returns a string describing the method which only contains the argument
 * names.  Basically the same thing as a selector name, except that it will
 * be prefixed with '+' or '-'.
"*/
- (NSString *)methodName
{
	NSString *selName = [self selectorName];
	return [NSString stringWithFormat:@"%@ %@", (isClassMethod)? @"+" : @"-", selName];
}

/*"
 * Returns a string describing the method, which will look a lot like the
 * original declaration -- it has the return type, argument types, variable
 * names, etc.
"*/
- (NSString *)fullMethodName
{
	NSMutableString *returnString = [NSMutableString stringWithCString:""];
	
	[returnString appendString:(isClassMethod)? @"+ " : @"- "];
	[returnString appendString:returnType];

	if ([self argumentCount] == 0)
	{
		[returnString appendString:[argumentNames objectAtIndex:0]];
	}
	else
	{
		int i, count = [self argumentCount];
		
		for (i=0;i<count;i++)
		{
			[returnString appendFormat:@"%@:%@%@ ",
				[argumentNames objectAtIndex:i],
				[argumentTypes objectAtIndex:i],
				[variableNames objectAtIndex:i]];
		}
	}
	
	return returnString;
}

- description
{
	return [NSString stringWithFormat:@"%@%@",
			[self entityDescription], [self fullMethodName]];
}

- (unsigned)argumentCount
{
	return [variableNames count];
}

/*"
 * Compares the receiver to the ADMethod otherMethod.  The comparison is
 * based on the -methodName of the two ADMethods.
"*/
- (NSComparisonResult)compareMethod:(ADMethod *)otherMethod
{
	return [[self methodName] compare:[otherMethod methodName]];
}

/*" Override to return the list of variable names "*/
- (NSArray *)wordsToItalicize
{
	if (!isVarargMethod)
		return variableNames;
	else
	{
		NSMutableArray *names = [variableNames mutableCopy];
		NSString *lastName = [[variableNames lastObject] substringToString:@","];
		unsigned lastIndex = [variableNames count] - 1;
		[names replaceObjectAtIndex:lastIndex withObject:lastName];
		return [names autorelease];
	}
}

@end
