//
//    ADInputHandler.m 
//      Written by Carl Lindberg Copyright 1999 by Carl Lindberg.
//                       All rights reserved.
//        This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
//  This file formatted with 4 spaces per tabstop.

#import "ADInputHandler.h"
#import <Foundation/NSException.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSArray.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "NSFileManager+Autodoc.h"
#import "ADDefaults.h"
#import "ADUtilities.h"
#import "ADFile.h"
#import "ADClass.h"
#import "ADOutputModule.h"

@implementation ADInputHandler
/*"
 * ADInputHandler takes over file processing once the main loop figures out
 * which ADInputHandler subclass is responsible for the file.  The
 * ADInputHandler subclass is responsible for generating all needed
 * documentation.
 * 
 * The main loop simply calls the method +#handleInputFile:, which all
 * subclasses must override.  There are several other support methods
 * implemented that subclasses can use.
 * 
 * For now, all methods are class methods, since I couldn't figure out why
 * we really needed instances.  I may change them all to be instance
 * methods, though, just in case ivars are useful for a subclass.
"*/

/*"
 * Generate documentation from the given filename.  This method must be
 * overridden by subclasses.
"*/
+ (void)handleInputFile:(NSString *)filename
{
	[NSException raise:NSInternalInconsistencyException format:@"%@ must implement +%@!",
		NSStringFromClass(self), NSStringFromSelector(_cmd)];
}

/*"
 * Returns the full output filename for a given input filename.  Takes into
 * account the -destination, -project, and -tree parameters, and will
 * return a file with the current ADOutputModule's documentation extension.
"*/
+ (NSString *)outputFilenameForInputFilename:(NSString *)inputFilename
{
	NSString		*projectDirectory = ADStringForDefault(ADProjectDirectory);
	BOOL			preserveTree      = ADBoolForDefault(ADPreserveTree);
	NSString		*outputDirectory  = ADStringForDefault(ADOutputDirectory);
	NSString		*outputFile       = nil;
	ADOutputModule	*outputModule     = [ADOutputModule currentModule];
	
    /*
     * Preserve a project's directory tree if desired.  This is only
     * applicable when parsing a project and an output directory has been
     * set (since otherwise documentation files are generated inside the
     * project directory).
     */
	if (projectDirectory && outputDirectory && preserveTree &&
		[inputFilename hasPrefix:projectDirectory])
	{
		NSString *relativePath = [inputFilename substringFromEndOfString:projectDirectory];
		outputFile = [outputDirectory stringByAppendingString:relativePath];
	}
	/* Otherwise, if an output directory was set, put the docs there. */
	else if (outputDirectory)
	{
		outputFile = [outputDirectory stringByAppendingPathComponent:
										[inputFilename lastPathComponent]];
	}
	/* Default to putting the docs in the same dir as the source. */
	else
	{
		outputFile = inputFilename;
	}

	outputFile = ADPathWithExtension(outputFile, [outputModule fileExtension]);

	return outputFile;
}

/*"
 * Returns the output filename for the given ADFile instance.  This is
 * similar to +#outputFilenameForInputFilename:, except that if the
 * -allclasses flag was given the +#uniqueFileBasename of adFile's
 * -#mainInterface is used for the file basename, and if -typedirs was
 * given, an extra directory path component is inserted under the
 * -destination directory.
"*/
+ (NSString *)outputFilenameForADFile:(ADFile *)adFile
{
	NSString *inputFile = [adFile headerFilename];
    NSString *outputFile;
	
	if (ADBoolForDefault(ADDocumentAllClasses) && [adFile hasMainInterface])
	{
		NSString *path = [[adFile headerFilename] stringByDeletingLastPathComponent];
		NSString *base = [[adFile mainInterface] uniqueFileBasename];
		
		inputFile = [path stringByAppendingPathComponent:base];
	}

    outputFile = [self outputFilenameForInputFilename:inputFile];

    /*
     * Handle the -typedirs command line parameter, and insert a a "type"
     * directory (Class, Protocols, Functions, or TypesAndConstants)
     * underneath the destination directory.  This parameter is dependant
     * on the destination directory being set and is mutally exclusive with
     * -tree, so make sure the the former is set and the latter isn't before
     * going ahead.
     */
    if (ADStringForDefault(ADOutputDirectory) != nil &&
		!ADBoolForDefault(ADPreserveTree) &&
        ADBoolForDefault(ADUseDocTypeDirectories))
    {
        NSString *filename = [outputFile lastPathComponent];
        NSString *base     = [outputFile stringByDeletingLastPathComponent];
        NSString *typeDir  = [adFile typeDirectoryName];

        outputFile = [[base stringByAppendingPathComponent:typeDir]
                        stringByAppendingPathComponent:filename];
    }

    return outputFile;
}

/*"
 * Returns YES if the targetFilename needs to be updated from
 * sourceFilename. This is true if sourceFilename is newer than
 * targetFilename or the -force flag was given.
"*/
+ (BOOL)shouldOutputToFile:(NSString *)targetFilename fromFile:(NSString *)sourceFilename
{
    if (ADBoolForDefault(ADForceOverwrite)) return YES;

	return ADFileIsNewer(sourceFilename, targetFilename);
}

/*"
 * The same as +#shouldOutputToFile:fromFile:, except that both adFile's
 * header and source filenames are checked.  Also will return NO if adFile
 * doesn't have anything to document.
"*/
+ (BOOL)shouldOutputToFile:(NSString *)targetFilename fromADFile:(ADFile *)adFile
{
    if ([adFile hasParsed] && ![adFile hasDocumentableElements])
    {
        ADDebugLog(1, @"Nothing to document in %@, skipping.", [adFile headerFilename]);
        return NO;
    }

	if ([self shouldOutputToFile:targetFilename fromFile:[adFile headerFilename]])
		return YES;
	
	if ([adFile sourceFilename] == nil)
		return NO;
	
	return [self shouldOutputToFile:targetFilename fromFile:[adFile sourceFilename]];
}

/*"
 * Creates the path to the given filename if needed, and generates
 * log/error messages as necessary.
"*/
+ (void)createPathForOutputFile:(NSString *)filename
{
	NSFileManager *fileManager = [NSFileManager defaultManager];
	NSString	  *outputDirectory = [filename stringByDeletingLastPathComponent];

	if ([outputDirectory length] > 0 && ![fileManager directoryExistsAtPath:outputDirectory])
	{
		BOOL beSilent = ADBoolForDefault(ADRunSilently);

		if (!beSilent) ADLog(@"Creating directory %@", outputDirectory);
		if (![fileManager deepCreateDirectoryAtPath:outputDirectory attributes:nil])
			ADLog(@"Failed to create directory %@", outputDirectory);
	}
}

/*"
 * A convenience method for generating a documentation file from an ADFile
 * instance. It checks to see if the output file needs to be created, and
 * if so, calls +#outputContents:toFile: on the current ADOutputModule.
"*/
+ (void)outputADFile:(ADFile *)adFile
{
	ADOutputModule *outputModule = [ADOutputModule currentModule];
	NSString       *outputFilename = [self outputFilenameForADFile:adFile];
	
	if ([self shouldOutputToFile:outputFilename fromADFile:adFile])
	{
		[self createPathForOutputFile:outputFilename];
		[outputModule outputContents:adFile toFile:outputFilename];
	}
}

/*"
 * A convenience method for generating documentation from an ADFile
 * instance. Simply calls +#outputADFile:, unless -allclasses was given, in
 * which case it calls +#outputADFile: for each class/protocol/category in
 * adFile.  Also handles coalescing of categories if -combine was given.
"*/
+ (void)handleMainADFile:(ADFile *)adFile
{
    if (ADBoolForDefault(ADCoalesceCategories))
        [adFile coalesceCategories];

	if (!ADBoolForDefault(ADDocumentAllClasses))
	{
		[self outputADFile:adFile];
	}
	else
	{
		NSEnumerator *fileEnum = [[adFile subfileArray] objectEnumerator];
		ADFile       *subFile;

		while ((subFile = [fileEnum nextObject]))
		{
			[self outputADFile:subFile];
		}
	}
}

@end
